@@ -19,22 +19,22 @@
 (macsyma-module outmis)
 
 
-(declare-top (special *xvar $exptisolate $labels $dispflag errorsw))
+(declare-top (special $exptisolate $labels $dispflag errorsw))
 
 (defmvar $exptisolate nil)
 (defmvar $isolate_wrt_times nil)
 
 (defmfun $isolate (e *xvar)
-  (setq *xvar (getopr *xvar)) (iso1 e))
+  (iso1 e (getopr *xvar)))
 
-(defun iso1 (e)
-  (cond ((specrepp e) (iso1 (specdisrep e)))
+(defun iso1 (e *xvar)
+  (cond ((specrepp e) (iso1 (specdisrep e) *xvar))
 	((and (free e 'mplus) (or (null $isolate_wrt_times) (free e 'mtimes))) e)
 	((freeof *xvar e) (mgen2 e))
 	((alike1 *xvar e) *xvar)
-	((member (caar e) '(mplus mtimes) :test #'eq) (iso2 e))
+	((member (caar e) '(mplus mtimes) :test #'eq) (iso2 e *xvar))
 	((eq (caar e) 'mexpt)
-	 (cond ((null (atom (cadr e))) (list (car e) (iso1 (cadr e)) (caddr e)))
+	 (cond ((null (atom (cadr e))) (list (car e) (iso1 (cadr e) *xvar) (caddr e)))
 	       ((or (alike1 (cadr e) *xvar) (not $exptisolate)) e)
 	       (t (let ((x ($rat (caddr e) *xvar)) (u 0) (h 0))
 		    (setq u (ratdisrep ($ratnumer x)) x (ratdisrep ($ratdenom x)))
@@ -42,18 +42,18 @@
 			(setq u ($multthru (list '(mexpt) x -1) u)))
 		    (if (mplusp u)
 			(setq u ($partition u *xvar) h (cadr u) u (caddr u)))
-		    (setq u (power* (cadr e) (iso1 u)))
+		    (setq u (power* (cadr e) (iso1 u *xvar)))
 		    (cond ((not (equal h 0))
 			   (mul2* (mgen2 (power* (cadr e) h)) u))
 			  (t u))))))
-	(t (cons (car e) (mapcar #'iso1 (cdr e))))))
+	(t (cons (car e) (mapcar #'(lambda (e1) (iso1 e1 *xvar)) (cdr e))))))
 
-(defun iso2 (e)
+(defun iso2 (e *xvar)
   (prog (hasit doesnt op)
      (setq op (ncons (caar e)))
      (do ((i (cdr e) (cdr i))) ((null i))
        (cond ((freeof *xvar (car i)) (setq doesnt (cons (car i) doesnt)))
-	     (t (setq hasit (cons (iso1 (car i)) hasit)))))
+	     (t (setq hasit (cons (iso1 (car i) *xvar) hasit)))))
      (cond ((null doesnt) (go ret))
 	   ((and (null (cdr doesnt)) (atom (car doesnt))) (go ret))
 	   ((prog2 (setq doesnt (simplify (cons op doesnt)))
@@ -93,8 +93,9 @@
 
 (defmfun $reveal (e lev)
   (setq e (format1 e))
-  (cond ((and (eq (ml-typep lev) 'fixnum) (> lev 0)) (reveal e 1 lev))
-	(t (merror (intl:gettext "reveal: second argument must be a positive integer; found: ~M") lev))))
+  (if (and (fixnump lev) (plusp lev))
+      (reveal e 1 lev)
+      (merror (intl:gettext "reveal: second argument must be a positive integer; found: ~M") lev)))
 
 (defun simple (x)
   (or (atom x) (member (caar x) '(rat bigfloat) :test #'eq)))
@@ -132,7 +133,7 @@
 (defun properties (x)
   (if (stringp x)
     ; AT THIS POINT WE MIGHT WANT TO TRY TO TEST ALL CHARS IN STRING ...
-    (if (and (> (length x) 0) (member (char x 0) *alphabet* :test #'eq))
+    (if (and (> (length x) 0) (member (char x 0) *alphabet*))
       '((mlist) $alphabetic)
       '((mlist)))
     (do ((y (symbol-plist x) (cddr y))
@@ -255,7 +256,7 @@
     (do ((ll (mget (car l) 'atvalues) (cdr ll)))
 	((null ll))
       (mtell-open "~M~%"
-		  (list '(mlable) nil
+		  (list '(mlabel) nil
 			(list '(mequal)
 			      (atdecode (car l) (caar ll) (cadar ll)) (caddar ll))))))
   '$done)
@@ -291,7 +292,7 @@
     (do ((j (mget (car i) '$atomgrad) (cdr j)))
 	((null j))
       (mtell-open "~M~%"
-		  (list '(mlable) nil
+		  (list '(mlabel) nil
 			(list '(mequal)
 			      (list '(%derivative) (car i) (caar j) 1) (cdar j))))))
   '$done)
@@ -305,7 +306,7 @@
 	 (thing (cons (ncons (car i)) (car l))))
 	((or (null k) (null j)))
       (mtell-open "~M~%"
-		  (list '(mlable)
+		  (list '(mlabel)
 			nil (list '(mequal) (list '(%derivative) thing (car j) 1.) (car k))))))
   '$done)
 
@@ -318,57 +319,107 @@
 			    (ncons (car i)))
 		    ret))))
 
-(declare-top (special trans ovar nvar tfun invfun $programmode nfun
-		      *roots *failures varlist genvar $ratfac))
+(declare-top (special $programmode *roots *failures varlist genvar $ratfac))
 
 (defmfun $changevar (expr trans nvar ovar)
-  (let (invfun nfun $ratfac)
-    (cond ((or (atom expr) (eq (caar expr) 'rat) (eq (caar expr) 'mrat))  expr)
-	  ((atom trans) (merror (intl:gettext "changevar: second argument must not be an atom; found: ~M") trans))
-	  ((null (atom nvar)) (merror (intl:gettext "changevar: third argument must be an atom; found: ~M") nvar))
-	  ((null (atom ovar)) (merror (intl:gettext "changevar: fourth argument must be an atom; found: ~M") ovar)))
-    (setq tfun (solvable (setq trans (meqhk trans)) ovar))
-    (changevar expr)))
+  (let ($ratfac)
+    (cond ((or (atom expr) (eq (caar expr) 'rat) (eq (caar expr) 'mrat))
+	   expr)
+	  ((atom trans)
+	   (merror (intl:gettext "changevar: second argument must not be an atom; found: ~M") trans))
+	  ((null (atom nvar))
+	   (merror (intl:gettext "changevar: third argument must be an atom; found: ~M") nvar))
+	  ((null (atom ovar))
+	   (merror (intl:gettext "changevar: fourth argument must be an atom; found: ~M") ovar)))
+    (changevar expr trans nvar ovar)))
 
 (defun solvable (l var &optional (errswitch nil))
   (let (*roots *failures)
     (solve l var 1)
-    (cond (*roots ($rhs (car *roots)))
+    (cond (*roots
+	   ;; We arbitrarily pick the first root.  Should we be more careful?
+	   ($rhs (car *roots)))
 	  (errswitch (merror (intl:gettext "changevar: failed to solve for ~M in ~M") var l))
 	  (t nil))))
 
-(defun changevar (expr)
+(defun changevar (expr trans nvar ovar)
   (cond ((atom expr) expr)
 	((or (not (member (caar expr) '(%integrate %sum %product) :test #'eq))
 	     (not (alike1 (caddr expr) ovar)))
-	 (recur-apply #'changevar expr))
-	(t (let ((deriv (if tfun (sdiff tfun nvar)
-			    (neg (div (sdiff trans nvar) ;IMPLICIT DIFF.
-				      (sdiff trans ovar))))))
-	     (cond ((and (member (caar expr) '(%sum %product) :test #'eq)
-			 (not (equal deriv 1)))
-		    (merror (intl:gettext "changevar: illegal change in summation or product")))
-		   ((setq nfun ($radcan	;NIL IF KERNSUBST FAILS
-				(if tfun
-				    (mul (maxima-substitute tfun ovar (cadr expr))
-					 deriv)
-				    (kernsubst ($ratsimp (mul (cadr expr)
-							      deriv))
-					       trans ovar))))
-		    (cond ;; DEFINITE INTEGRAL,SUMMATION, OR PRODUCT
-		      ((cdddr expr)
-		       (or invfun (setq invfun (solvable trans nvar t)))
-		       (list (ncons (caar expr)) ;THIS WAS CHANGED
-			     nfun	;FROM '(%INTEGRATE)
-			     nvar
-			     ($limit invfun ovar (cadddr expr) '$plus)
-			     ($limit invfun
-				     ovar
-				     (car (cddddr expr))
-				     '$minus)))
-		      (t		;INDEFINITE INTEGRAL
-		       (list '(%integrate) nfun nvar))))
-		   (t expr))))))
+	 (recur-apply (lambda (e) (changevar e trans nvar ovar)) expr))
+	(t
+	 ;; TRANS is the expression that relates old var and new var
+	 ;; and is of the form f(ovar, nvar) = 0. Using TRANS, try to
+	 ;; solve for ovar so that ovar = tfun(nvar), if possible.
+	 (let* ((tfun (solvable (setq trans (meqhk trans)) ovar))
+		(deriv
+		 ;; Compute diff(tfun, nvar) = dovar/dnvar if tfun is
+		 ;; available.  Otherwise, use implicit
+		 ;; differentiation.
+		 (if tfun
+		     (sdiff tfun nvar)
+		     (neg (div (sdiff trans nvar) ;IMPLICIT DIFF.
+			       (sdiff trans ovar)))))
+		(sum-product-p (member (caar expr) '(%sum %product) :test #'eq)))
+
+	   #+nil
+	   (progn
+	     (mformat t "tfun = ~M~%" tfun)
+	     (mformat t "deriv = ~M~%" deriv))
+
+	   ;; For sums and products, we want deriv to be +/-1 because
+	   ;; I think that means that integers will map into integers
+	   ;; (roughly), so that we don't need to express the
+	   ;; summation index or limits in some special way to account
+	   ;; for it.
+	   (when (and (member (caar expr) '(%sum %product) :test #'eq)
+		      (not (or (equal deriv 1)
+			       (equal deriv -1))))
+	     (merror (intl:gettext "changevar: illegal change in summation or product")))
+
+	   (let ((nfun ($radcan	;NIL IF KERNSUBST FAILS
+			      (if tfun
+				  (mul (maxima-substitute tfun ovar (cadr expr))
+				       ;; Don't multiply by deriv
+				       ;; for sums/products because
+				       ;; reversing the order of
+				       ;; limits doesn't change the
+				       ;; sign of the result.
+				       (if sum-product-p 1 deriv))
+				  (kernsubst ($ratsimp (mul (cadr expr)
+							    deriv))
+					     trans ovar)))))
+	     (cond
+	       (nfun
+		;; nfun is basically the result of subtituting ovar
+		;; with tfun in the integratand (summand).
+		(cond	
+		  ((cdddr expr)
+		   ;; Handle definite integral, summation, or product.
+		   ;; invfun expresses nvar in terms of ovar so that
+		   ;; we can compute the new lower and upper limits of
+		   ;; the integral (sum).
+		   (let* ((invfun (solvable trans nvar t))
+			  (lo-limit ($limit invfun ovar (cadddr expr) '$plus))
+			  (hi-limit ($limit invfun
+					    ovar
+					    (car (cddddr expr))
+					    '$minus)))
+		     ;; If this is a sum or product and deriv = -1, we
+		     ;; want to reverse the low and high limits.
+		     (when (and sum-product-p (equal deriv -1))
+		       (rotatef lo-limit hi-limit))
+
+		     ;; Construct the new result.
+		     (list (ncons (caar expr))
+			   nfun
+			   nvar
+			   lo-limit
+			   hi-limit)))
+		  (t
+		   ;; Indefinite integral
+		   (list '(%integrate) nfun nvar))))
+	       (t expr)))))))
 
 (defun kernsubst (expr form ovar)
   (let (varlist genvar nvarlist)
